
package shopping.cart

import akka.Done
import akka.NotUsed
import akka.actor.typed.ActorSystem
import akka.kafka.ProducerMessage
import akka.kafka.ProducerSettings
import akka.kafka.scaladsl.Producer
import akka.projection.eventsourced.EventEnvelope
import akka.projection.ProjectionContext
import akka.stream.scaladsl.FlowWithContext
import com.google.protobuf.any.{ Any => ScalaPBAny }
import org.apache.kafka.clients.producer.ProducerRecord
import org.apache.kafka.common.serialization.ByteArraySerializer
import org.apache.kafka.common.serialization.StringSerializer
import org.slf4j.LoggerFactory

object PublishEventsProjectionFlow {
  private val log = LoggerFactory.getLogger(getClass)

  def createProducerFlow(
      topic: String,
      system: ActorSystem[_]): FlowWithContext[
    EventEnvelope[ShoppingCart.Event],
    ProjectionContext,
    Done,
    ProjectionContext,
    NotUsed] = {

    val producerSettings =
      ProducerSettings(system, new StringSerializer, new ByteArraySerializer)

    
    FlowWithContext[EventEnvelope[ShoppingCart.Event], ProjectionContext]
      .map(envelope => {
        // using the cartId as the key and `DefaultPartitioner` will select partition based on the key
        // so that events for same cart always ends up in same partition
        val key = envelope.event.cartId
        val record =
          new ProducerRecord(topic, key, serialize(envelope.event)) 
        log.info("Publishing event [{}] to topic {}", envelope.event, topic)
        ProducerMessage.single(record)
      })
      .via(Producer.flowWithContext(producerSettings))
      .map(_ => Done)
  }

  private def serialize(event: ShoppingCart.Event): Array[Byte] = {
    val protoMessage = event match {
      case ShoppingCart.ItemAdded(cartId, itemId, quantity) =>
        proto.ItemAdded(cartId, itemId, quantity)
      case ShoppingCart.CheckedOut(cartId, _) =>
        proto.CheckedOut(cartId)
      
      case ShoppingCart.ItemQuantityAdjusted(cartId, itemId, newQuantity, _) =>
        proto.ItemQuantityAdjusted(cartId, itemId, newQuantity)
      case ShoppingCart.ItemRemoved(cartId, itemId, _) =>
        proto.ItemRemoved(cartId, itemId)
      
    }
    // pack in Any so that type information is included for deserialization
    ScalaPBAny.pack(protoMessage).toByteArray 
  }
}

